// Object that needs to be exported for getting the ouptut parameters
export let deviceConfig = {};

// ---------- Controls & Constants ----------

deviceConfig.fclk_div = 8;
// Get tmod delay count
let DELAY_LIST = [
  0, 46, 54, 62, 78, 110, 174, 302, 558, 1070, 2094, 4142, 8238, 16430,
];
// Get tmod converion count
// WARNING 1: These numbers were rounded and may off by a few mod clK periods
// WARNING 2: I'm not sure of the device behavior when the FIR filter is selected but a non-FIR data rate is programmed...
let SETTLING_LIST = [
  [
    369055, 184735, 92575, 55738, 46522, 18874, 15796, 9658, 2742, 1205, 821,
    629, 565, 435, 355, 295, 211,
  ], // UPDATE!
  [
    737695, 369055, 184735, 111007, 92575, 37306, 31159, 18874, 5046, 1973,
    1205, 821, 693, 435, 355, 295, 211,
  ],
  [
    1105966, 553375, 276895, 166303, 138655, 55738, 46522, 28090, 7350, 2742,
    1590, 1014, 821, 435, 355, 295, 211,
  ],
  [
    1474606, 737695, 369055, 221599, 184735, 74170, 61876, 37306, 9658, 3509,
    1973, 1205, 949, 435, 355, 295, 211,
  ],
  [370714, 186348, 94180, 0, 48103, 0, 0, 0, 0, 0, 0, 0, 0, 435, 355, 295, 211],
];

// ---------- Helper Methods ---------------

// Function that needs to be exported for getting the filter response (arguments should be in the same order and it will be passed from the GUI)
export function getFilterResponse(filter, dataRate, fclk, chop, lutData) {
  deviceConfig.fmod = fclk / deviceConfig.fclk_div;
  let SINC5_OSR_LIST = [
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 48, 36, 32,
  ];
  let SINCx_OSR_LIST = [
    5760, 2880, 1440, 864, 720, 288, 240, 144, 36, 12, 6, 3, 2, 1, 1, 1, 1,
  ];
  let AVG_OSR_LIST = [8, 4, 2, 1];
  let SINCx_LIST = ["SINC1", "SINC2", "SINC3", "SINC4", "SINC1"];

  let SINC5_OSR = SINC5_OSR_LIST[dataRate];
  let SINCx_OSR = SINCx_OSR_LIST[dataRate];
  let SINCx = SINCx_LIST[filter];

  let IS_FIR_DATA_RATE = [0, 1, 2, 4].includes(dataRate);

  let STAGES;
  let OSR;
  if (dataRate >= 13) {
    // >= 14.4 kSPS
    STAGES = ["SINC5"];
    OSR = [SINC5_OSR];
  } else if (filter !== 4 || !IS_FIR_DATA_RATE) {
    // SINCx
    STAGES = ["SINC5", SINCx];
    OSR = [SINC5_OSR, SINCx_OSR];
  } else if (dataRate === 4) {
    // FIR (20 SPS)
    STAGES = ["SINC5", "SINC1", "FIR"];
    OSR = [SINC5_OSR, 24, 30];
  } else {
    STAGES = ["SINC5", "SINC1", "FIR", "SINC1"];
    OSR = [SINC5_OSR, 24, 30, AVG_OSR_LIST[dataRate]];
  }

  if (chop === 1) {
    STAGES.push("CHOP"); // adding CHOP to the stages if chop is enabled
  }

  // Compute output data rate
  // WITHOUT CHOP: Divide fmod by combined OSR...
  if (chop === 0) {
    deviceConfig.data_rate_sps =
      deviceConfig.fmod /
      OSR.reduce((accumulator, currentValue) => accumulator * currentValue); // product of OSR values
  }

  // WITH CHOP: Output data rate equals the filter settling time (without chop)
  else if (chop === 1) {
    //Calculating filter settling time since it is required for the data rate calculation
    deviceConfig.tMOD_ms = 1000 / deviceConfig.fmod;
    deviceConfig.tMOD_CONVERSION_COUNT =
      deviceConfig.tMOD_DELAY_COUNT + SETTLING_LIST[filter][dataRate]; // (row index, column index)
    deviceConfig.single_conversion_time_ms =
      deviceConfig.tMOD_CONVERSION_COUNT * deviceConfig.tMOD_ms; // For all conversions

    deviceConfig.data_rate_sps = 1000 / deviceConfig.single_conversion_time_ms; // Storing ouput data rate value in device config since it is required in the GUI
  }

  // Adjust for fclk frequency...
  deviceConfig.fclk_nom = 7.3728e6; // storing nominal frequency value in device config since it is required in the GUI
  // LUT.Freq * (fclk/fclk_nom)
  deviceConfig.x_freqs_adj = lutData.Freq.map(
    num => num * (fclk / deviceConfig.fclk_nom),
  ); // storing adjusted frequency value in device config since it is required in the GUI

  // Loop through 'columnsToSum' array, computing response from LUT...
  let columnsToSum = STAGES;
  let filter_response = Array(deviceConfig.x_freqs_adj.length).fill(0); // filling the x_freqs_adj array with 0
  for (let i = 0; i < columnsToSum.length; i++) {
    // Get filter type and input fs
    let filterType = columnsToSum[i];
    let intermediate_osr = OSR.slice(0, i).reduce(
      (accumulator, currentValue) => accumulator * currentValue,
      1,
    ); // partial product of OSR values
    let fs = deviceConfig.fmod / intermediate_osr;

    // Compute SINCx filter?
    switch (filterType) {
      case "SINC1":
        lutData.SINC1 = computeSincxResponse(
          1,
          OSR[i],
          fs,
          deviceConfig.x_freqs_adj,
        );
        break;

      case "SINC2":
        lutData.SINC2 = computeSincxResponse(
          2,
          OSR[i],
          fs,
          deviceConfig.x_freqs_adj,
        );
        break;

      case "SINC3":
        lutData.SINC3 = computeSincxResponse(
          3,
          OSR[i],
          fs,
          deviceConfig.x_freqs_adj,
        );
        break;

      case "SINC4":
        lutData.SINC4 = computeSincxResponse(
          4,
          OSR[i],
          fs,
          deviceConfig.x_freqs_adj,
        );
        break;

      case "SINC5":
        lutData.SINC5 = computeSincxResponse(
          5,
          OSR[i],
          fs,
          deviceConfig.x_freqs_adj,
        );
        break;

      case "CHOP":
        lutData.CHOP = computeSincxResponse(1, 2, fs, deviceConfig.x_freqs_adj);
        break;
    }

    // Sum responses
    let column = lutData[filterType];
    // Both the column abd filter_response array should be of same length
    if (column.length != filter_response.length) {
      console.error(
        "Filter Response and Column arrays must have same length for element-wise addition",
      );
    }
    filter_response = filter_response.map((num, index) => num + column[index]); // adding filter_response and column array
  }
  return filter_response; // returning the calculated filter response
}

// Function that needs to be exported for getting the time to conversion (arguments should be in the same order and it will be passed from the GUI)
export function getFilterSettlingTime(filter, dataRate, fclk, chop, delay) {
  // Determine modular sampling rate and period
  deviceConfig.fmod = fclk / deviceConfig.fclk_div;
  deviceConfig.tMOD_ms = 1000 / deviceConfig.fmod;

  deviceConfig.tMOD_DELAY_COUNT = DELAY_LIST[delay];

  deviceConfig.tMOD_CONVERSION_COUNT =
    deviceConfig.tMOD_DELAY_COUNT + SETTLING_LIST[filter][dataRate]; // (row index, column index)

  // Compute times...
  deviceConfig.conv_start_delay_ms =
    deviceConfig.tMOD_ms * deviceConfig.tMOD_DELAY_COUNT;
  deviceConfig.single_conversion_time_ms =
    deviceConfig.tMOD_CONVERSION_COUNT * deviceConfig.tMOD_ms; // For all conversions
  deviceConfig.filter_settling_time_ms =
    (chop + 1) * deviceConfig.single_conversion_time_ms; // First conversion only

  return deviceConfig.filter_settling_time_ms; // returning the calculated filter settling time
}

// ---------- Utilies ---------------

// Helper function for calculating the sinc response
function computeSincxResponse(order, osr, fs, x_freqs) {
  let sincx_response = Array(x_freqs.length).fill(0); // filling the x_freqs array with 0
  // 0th index is not considered in the loop, since the value will be zero
  for (let i = 1; i < x_freqs.length; i++) {
    let fin = x_freqs[i];
    let arg = (Math.PI * fin) / fs;
    let ratio = Math.abs(Math.sin(arg * osr) / (osr * Math.sin(arg)));
    sincx_response[i] = 20 * Math.log10(Math.abs(ratio) ** order);
  }
  return sincx_response; //returning the calculated sinc response
}
